/* -*- mode:c -*- */
/** @file
 ** @brief Maestro NUMA abstraction
 **/
/*
 * Copyright (C) 2021 HPE Switzerland GmbH
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MAESTRO_I_NUMA_H_
#define MAESTRO_I_NUMA_H_ 1

/**@addtogroup MSTRO_Internal
 **@{
 **/

/**@defgroup MSTRO_I_NUMA Maestro NUMA abstraction
 **@{
 **
 **/

#ifdef HAVE_NUMAIF_H
#include <numa.h>
#include <numaif.h>
#endif

#ifdef __APPLE__
#include <cpuid.h>
#endif

#include "maestro/logging.h"


/** obtain node ID for a given address.
 *
 * Will always succeed, yielding '0' if NUMA is not supported or ID cannot be obtained */
static inline
int
mstro_numa__node_for_addr(void * addr)
{
#ifdef HAVE_NUMA
  if(g_have_numa) {
    int node = -1;
    long s = get_mempolicy(&node,
		    NULL /* node mask */,
		    0 /* max nodes in mask */,
		    addr, MPOL_F_NODE | MPOL_F_ADDR);
    if(s<0) {
      LOG_WARN(MSTRO_LOG_MODULE_CORE,
	       "Failed to obtain NUMA info for addr %p: %ld (%s)\n",
	       addr, s, strerror(-s));
      return 0;
    } else {
      return node;
    }
  } else {
    return 0;
  }
#else
  void *dummy = addr; /* avoid 'unused arg' warning */
  return 0;
#endif
}

/** Return the number of NUMA domains available to this maestro instance */
static inline
int
mstro_numa__num_nodes(void)
{
#ifdef HAVE_NUMA
  /* FIXME: should query resource/config info */
  return numa_max_node()+1;
#else
  return 1;
#endif
}

static inline 
int
mstro_numa_get_cpu(void) {
  #if defined(__linux) || defined(__linux__)
  return sched_getcpu();
  #elif __APPLE__
  #define CPUID(INFO, LEAF, SUBLEAF) __cpuid_count(LEAF, SUBLEAF, INFO[0], INFO[1], INFO[2], INFO[3])
      uint32_t CPUInfo[4];                           
      CPUID(CPUInfo, 1, 0);                          
       /* CPUInfo[1] is EBX, bits 24-31 are APIC ID */ 
      if ( (CPUInfo[3] & (1 << 9)) == 0) {           
          return  -1;  /* no APIC on chip */             
      }                                              
      int CPU = (unsigned)CPUInfo[1] >> 24;                                      
      return (CPU>=0)?CPU:0; 
  #else
  return 0;
  #endif

}

/**@} (end of group MSTRO_I_NUMA) */
/**@} (end of group MSTRO_Internal) */

#endif /* MAESTRO_I_NUMA_H_ */

